/*
 * Copyright (c) 1997, 1998  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#include "eb.h"
#include "error.h"
#include "io.h"
#include "language.h"

#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif /* HAVE_STRCHR */

#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifndef HAVE_MEMCPY
#define memcpy(d, s, n) bcopy((s), (d), (n))
#ifdef __STDC__
void *memchr(const void *, int, size_t);
int memcmp(const void *, const void *, size_t);
void *memmove(void *, const void *, size_t);
void *memset(void *, int, size_t);
#else /* not __STDC__ */
char *memchr();
int memcmp();
char *memmove();
char *memset();
#endif /* not __STDC__ */
#endif

/*
 * Whence parameter for lseek().
 */
#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

/*
 * The maximum length of path name.
 */
#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX	MAXPATHLEN
#else /* not MAXPATHLEN */
#define PATH_MAX	1024
#endif /* not MAXPATHLEN */
#endif /* not PATH_MAX */

/*
 * Unexported functions.
 */
#ifdef __STDC__
static int eb_initialize_language(EB_Book *);
static int eb_initialize_catalog(EB_Book *);
static int eb_initialize_indexes(EB_Book *);
static int eb_initialize_eb_fonts(EB_Book *);
static int eb_initialize_epwing_fonts(EB_Book *);
#else /* not __STDC__ */
static int eb_initialize_language();
static int eb_initialize_catalog();
static int eb_initialize_indexes();
static int eb_initialize_eb_fonts();
static int eb_initialize_epwing_fonts();
#endif /* not __STDC__ */


/*
 * Unexported variable.
 * Book ID counter.
 */
static EB_Book_Code counter = 0;


/*
 * Initialize `book'.
 */
void
eb_initialize(book)
    EB_Book *book;
{
    book->code = counter++;
    book->path = NULL;
    book->subcurrent = NULL;
    book->subbooks = NULL;
    book->languages = NULL;
    book->langcurrent = NULL;
    book->messages = NULL;
}


/*
 * Bind `book' to `path'.
 */
int
eb_bind(book, path)
    EB_Book *book;
    const char *path;
{
    char tmppath[PATH_MAX + 1];
    int count;

    /*
     * Reset structure members in `book'.
     */
    eb_initialize(book);

    /*
     * Set the path of the appendix.
     * The length of the filename "path/subdir/subsubdir/file.;1" must
     * not exceed PATH_MAX.
     */
    if (PATH_MAX < strlen(path)) {
	eb_error = EB_ERR_TOO_LONG_FNAME;
	goto failed;
    }
    strcpy(tmppath, path);
    if (eb_canonicalize_filename(tmppath) < 0)
	goto failed;

    book->pathlen = strlen(tmppath);
    if (PATH_MAX < book->pathlen + (1 + EB_MAXLEN_BASENAME + 1
	+ EB_MAXLEN_BASENAME + 1 + EB_MAXLEN_BASENAME + 3)) {
	eb_error = EB_ERR_TOO_LONG_FNAME;
	goto failed;
    }

    book->path = (char *)malloc(book->pathlen + 1);
    if (book->path == NULL) {
	eb_error = EB_ERR_MEMORY_EXHAUSTED;
	goto failed;
    }
    strcpy(book->path, tmppath);

    /*
     * Get disc type and filename mode.
     */
    if (eb_catalog_filename(book) < 0)
	goto failed;

    /*
     * Read information from the `LANGUAGES' file.
     * If failed to initialize, JISX0208 is assumed.
     */
    if (eb_initialize_language(book) < 0) {
	book->charcode = EB_CHAR_JISX0208;
	eb_error = EB_NO_ERR;
    }

    /*
     * Read information from the `CATALOG(S)' file.
     */
    count = eb_initialize_catalog(book);
    if (count < 0)
	goto failed;
    
    return count;

    /*
     * An error occurs...
     */
  failed:
    eb_clear(book);
    return -1;
}


/*
 * Suspend using `book'.
 */
void
eb_suspend(book)
    EB_Book *book;
{
    eb_unset_subbook(book);
    eb_unset_language(book);
}


/*
 * Finish using `book'.
 */
void
eb_clear(book)
    EB_Book *book;
{
    eb_unset_subbook(book);
    eb_unset_language(book);

    if (book->languages != NULL)
	free(book->languages);

    if (book->subbooks != NULL)
	free(book->subbooks);

    if (book->messages != NULL)
	free(book->messages);

    if (book->path != NULL)
	free(book->path);

    eb_initialize(book);
}


/*
 * Read information from the `LANGUAGES' file in `book'.
 *
 * If succeeded, the number of languages in the book is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
static int
eb_initialize_language(book)
    EB_Book *book;
{
    EB_Language *lng;
    int file = -1;
    int i;
    char language[PATH_MAX + 1];
    char buf[EB_SIZE_PAGE];
    char *bufp;

    /*
     * Open the language file.
     */
    sprintf(language, "%s/%s", book->path, EB_FNAME_LANGUAGE);
    eb_fix_filename(book, language);
    file = open(language, O_RDONLY | O_BINARY);
    if (file < 0) {
	eb_error = EB_ERR_FAIL_OPEN_LANG;
	goto failed;
    }

    /*
     * Get a character code of the book, and get the number of langueages
     * in the file.
     */
    if (read(file, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_LANG;
	goto failed;
    }

    book->charcode = eb_uint2(buf);
    if (book->charcode != EB_CHAR_ISO8859_1
	&& book->charcode !=  EB_CHAR_JISX0208) {
	eb_error = EB_ERR_UNEXP_LANG;
	goto failed;
    }

    book->langcount = eb_uint2(buf + 2);
    if (EB_MAX_LANGUAGES < book->langcount)
	book->langcount = EB_MAX_LANGUAGES;
    if (book->langcount == 0) {
	eb_error = EB_ERR_UNEXP_LANG;
	goto failed;
    }

    /*
     * Allocate memories for language entries.
     */
    book->languages = (EB_Language *) malloc(sizeof(EB_Language)
	* book->langcount);
    if (book->languages == NULL) {
	eb_error = EB_ERR_MEMORY_EXHAUSTED;
	goto failed;
    }

    /*
     * Get languege names.
     */
    if (read(file, buf, (EB_MAXLEN_LANGNAME + 1) * book->langcount)
	!= (EB_MAXLEN_LANGNAME + 1) * book->langcount) {
	eb_error = EB_ERR_FAIL_READ_LANG;
	goto failed;
    }
    for (i = 0, bufp = buf, lng = book->languages;
	 i < book->langcount; i++, bufp += EB_MAXLEN_LANGNAME + 1, lng++) {
	lng->code = eb_uint1(bufp);
	lng->offset = 0;
	lng->msgcount = 0;
	memcpy(lng->name, bufp + 1, EB_MAXLEN_LANGNAME);
	*(lng->name + EB_MAXLEN_LANGNAME) = '\0';
    }

    close(file);
    return book->langcount;

    /*
     * An error occurs...
     */
  failed:
    if (0 <= file)
	close(file);

    if (book->languages != NULL) {
	free(book->languages);
	book->languages = NULL;
    }

    return -1;
}


/*
 * Read information from the `LANGUAGES' file in `book'.
 *
 * If succeeded, the number of messages in the book is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
int
eb_initialize_messages(book)
    EB_Book *book;
{
    EB_Language *lng;
    int file = -1;
    ssize_t len;
    off_t offset;
    int maxmsgs;
    int i;
    char language[PATH_MAX + 1];
    char buf[EB_SIZE_PAGE];
    char *bufp;

    /*
     * Open the language file.
     */
    sprintf(language, "%s/%s", book->path, EB_FNAME_LANGUAGE);
    eb_fix_filename(book, language);
    file = open(language, O_RDONLY | O_BINARY);
    if (file < 0) {
	eb_error = EB_ERR_FAIL_OPEN_LANG;
	goto failed;
    }

    /*
     * Get a character code of the book, and get the number of langueages
     * in the file.
     */
    if (lseek(file, (EB_MAXLEN_LANGNAME + 1) * book->langcount + 16,
	SEEK_SET) < 0) {
	eb_error = EB_ERR_FAIL_SEEK_LANG;
	goto failed;
    }

    /*
     * Get offset and size of each language.
     */
    len = read(file, buf, EB_SIZE_PAGE);
    if (len < EB_MAXLEN_MESSAGE + 1) {
	eb_error = EB_ERR_UNEXP_LANG;
	goto failed;
    }
    if (*buf != book->languages->code || strncmp(buf + 1,
	book->languages->name, EB_MAXLEN_LANGNAME + 1) != 0) {
	eb_error = EB_ERR_UNEXP_LANG;
	goto failed;
    }

    i = 0;
    lng = book->languages;
    bufp = buf + EB_MAXLEN_LANGNAME + 1;
    offset = 16 + (EB_MAXLEN_LANGNAME + 1) * book->langcount
	+ (EB_MAXLEN_LANGNAME + 1);
    lng->offset = offset;

    for (;;) {
	int rest;
	EB_Language_Code code;

	/*
         * If `bufp' reached to a near of the end of the buffer,
	 * read next page.
         */
        rest = len - (bufp - buf);
        if (rest < EB_MAXLEN_MESSAGE + 1) {
	    int n;

            memcpy(buf, bufp, rest);
            n = read(file, buf + rest, EB_SIZE_PAGE - rest);
            if (n < 0 || rest + n < EB_MAXLEN_MESSAGE + 1)
                break;
	    len = n + rest;
	    bufp = buf;
        }

	code = eb_uint1(bufp);
	if (i + 1 < book->langcount && code == (lng + 1)->code
	    && strncmp((lng + 1)->name, bufp + 1, EB_MAXLEN_LANGNAME + 1)
	    == 0) {
	    /*
	     * Next language name is found.  I
	     */
	    i++;
	    lng++;
	    bufp += EB_MAXLEN_LANGNAME + 1;
	    offset += EB_MAXLEN_LANGNAME + 1;
	    lng->offset = offset;
	    lng->msgcount = 0;
	} else if (code == 0 && *(bufp + 1) == '\0') {
	    /*
	     * Break at an empty message.
	     */
	    break;
	} else {
	    /*
	     * A message.
	     */
	    bufp += EB_MAXLEN_MESSAGE + 1;
	    offset += EB_MAXLEN_MESSAGE + 1;
	    lng->msgcount++;
	}
    }

    /*
     * Truncate the number of messages in a language if exceeded its limit.
     */
    for (i = 0, lng = book->languages; i < book->langcount; i++, lng++) {
	if (EB_MAX_MESSAGES < lng->msgcount)
	    lng->msgcount = EB_MAX_MESSAGES;
    }

    /*
     * Decide the size for the message buffer, and allocate memories.
     */
    maxmsgs = 0;
    for (i = 0, lng = book->languages; i < book->langcount; i++, lng++) {
	if (maxmsgs < lng->msgcount)
	    maxmsgs = lng->msgcount;
    }
    if (maxmsgs == 0)
	maxmsgs = 1;

    book->messages = (char *)malloc((EB_MAXLEN_MESSAGE + 2) * maxmsgs);
    if (book->messages == NULL) {
	eb_error = EB_ERR_MEMORY_EXHAUSTED;
	goto failed;
    }

    /*
     * Close the language file.
     */
    close(file);

    return book->langcount;

    /*
     * An error occurs...
     */
  failed:
    if (0 <= file)
	close(file);

    if (book->messages != NULL) {
	free(book->messages);
	book->messages = NULL;
    }

    return -1;
}


/*
 * Read information from the `CATALOG(S)' file in 'book'.
 *
 * If succeeded, the number of subbooks in the book is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
static int
eb_initialize_catalog(book)
    EB_Book *book;
{
    char buf[EB_SIZE_PAGE];
    char catalog[PATH_MAX + 1];
    char *space;
    EB_Subbook *sub;
    size_t catalogsize;
    size_t titlesize;
    int file = -1;
    int i;

    if (book->disccode == EB_DISC_EB) {
	catalogsize = EB_SIZE_EB_CATALOG;
	titlesize = EB_MAXLEN_EB_TITLE;
    } else {
	catalogsize = EB_SIZE_EP_CATALOG;
	titlesize = EB_MAXLEN_EP_TITLE;
    }
	
    /*
     * Open the catalog file.
     */
    if (book->disccode == EB_DISC_EB)
	sprintf(catalog, "%s/%s", book->path, EB_FNAME_CATALOG);
    else
	sprintf(catalog, "%s/%s", book->path, EB_FNAME_CATALOGS);
    eb_fix_filename(book, catalog);
    file = open(catalog, O_RDONLY | O_BINARY);
    if (file < 0)
	goto failed;

    /*
     * Get the number of subbooks in this book.
     */
    if (read(file, buf, 16) != 16) {
	eb_error = EB_ERR_FAIL_READ_CAT;
	goto failed;
    }
    book->subcount = eb_uint2(buf);
    if (EB_MAX_SUBBOOKS < book->subcount)
	book->subcount = EB_MAX_SUBBOOKS;

    /*
     * Allocate memories for subbook entries.
     */
    book->subbooks = (EB_Subbook *) malloc(sizeof(EB_Subbook)
	* book->subcount);
    if (book->subbooks == NULL) {
	eb_error = EB_ERR_MEMORY_EXHAUSTED;
	goto failed;
    }

    /*
     * Read information about subbook.
     */
    for (i = 0, sub = book->subbooks; i < book->subcount; i++, sub++) {
	/*
	 * Read data from the catalog file.
	 */
	if (read(file, buf, catalogsize) != catalogsize) {
	    eb_error = EB_ERR_FAIL_READ_CAT;
	    goto failed;
	}

	/*
	 * Set a directory name.
	 */
	strncpy(sub->directory, buf + 2 + titlesize, EB_MAXLEN_BASENAME);
	sub->directory[EB_MAXLEN_BASENAME] = '\0';
	space = strchr(sub->directory, ' ');
	if (space != NULL)
	    *space = '\0';

	/*
	 * Set a title.  (Convert from JISX0208 to EUC JP)
	 */
	strncpy(sub->title, buf + 2, titlesize);
	sub->title[titlesize] = '\0';
	if (book->charcode == EB_CHAR_JISX0208)
	    eb_jisx0208_to_euc(sub->title, sub->title);

	/*
	 * If the book is EPWING, get font filenames.
	 */
	if (book->disccode == EB_DISC_EPWING) {
	    EB_Font *fnt = sub->fonts;
	    char *bufp;
	    int j;
	    int count = 0;

	    for (j = 0, bufp = buf + 2 + titlesize + 18;
		 j < EB_MAX_FONTS * 2; j++, bufp += EB_MAXLEN_BASENAME) {
		if (*bufp == '\0')
		    continue;
		strncpy(fnt->filename, bufp, EB_MAXLEN_BASENAME);
		fnt->filename[EB_MAXLEN_BASENAME] = '\0';
		space = strchr(fnt->filename, ' ');
		if (space != NULL)
		    *space = '\0';
		fnt++;
		count++;
	    }
	    sub->fontcount = count;
	}

	sub->initialized = 0;
	sub->code = i;
    }

    /*
     * Close the catalog file.
     */
    close(file);
    return book->subcount;

    /*
     * An error occurs...
     */
  failed:
    if (0 <= file)
	close(file);

    if (book->subbooks != NULL) {
	free(book->subbooks);
	book->subbooks = NULL;
    }

    return -1;
}


/*
 * Get information about the current subbook.
 */
int
eb_initialize_subbook(book)
    EB_Book *book;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Initialize search and text-output status.
     */
    eb_initialize_search();
    eb_initialize_text();

    /*
     * If the subbook has already initialized, return immediately.
     */
    if (book->subcurrent->initialized != 0)
	return 0;

    /*
     * Initialize members in the strucure EB_Subbook.
     */
    book->subcurrent->narwcurrent = NULL;
    book->subcurrent->widecurrent = NULL;
    if (book->disccode == EB_DISC_EB)
	book->subcurrent->fontcount = 0;

    /*
     * Read index information.
     */
    if (eb_initialize_indexes(book) < 0)
	return -1;

    /*
     * Read font information.
     */
    if (book->disccode == EB_DISC_EB) {
	if (eb_initialize_eb_fonts(book) < 0)
	    return -1;
    } else {
	if (eb_initialize_epwing_fonts(book) < 0)
	    return -1;
    }
    book->subcurrent->initialized = 1;

    /*
     * Rewind the file descriptor of the start file.
     */
    if (lseek(book->subcurrent->startfile, 0, SEEK_SET) < 0) {
	eb_error = EB_ERR_FAIL_SEEK_START;
	return -1;
    }

    return 0;
}


/*
 * Get information about all subbooks in the book.
 */
int
eb_initialize_all_subbooks(book)
    EB_Book *book;
{
    EB_Subbook_Code cur;
    EB_Subbook *sub;
    int i;

    /*
     * Get the current subbook.
     */
    if (book->subcurrent != NULL)
	cur = book->subcurrent->code;
    else
	cur = -1;

    /*
     * Initialize each subbook.
     */
    for (i = 0, sub = book->subbooks; i < book->subcount; i++, sub++) {
	if (eb_set_subbook(book, sub->code) < 0)
	    return -1;
    }

    /*
     * Restore the current subbook.
     */
    if (cur < 0)
	eb_unset_subbook(book);
    else if (eb_set_subbook(book, cur) < 0)
	return -1;

    return 0;
}


/*
 * Get index information in the current subbook.
 *
 * If succeeded, the number of indexes is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
static int
eb_initialize_indexes(book)
    EB_Book *book;
{
    EB_Subbook *sub = book->subcurrent;
    char buf[EB_SIZE_PAGE];
    char *bufp;
    int id;
    int count;
    int page;
    int i;

    /*
     * Read the index table in the subbook.
     */
    if (lseek(sub->startfile, 0, SEEK_SET) < 0) {
	eb_error = EB_ERR_FAIL_SEEK_START;
	return -1;
    }
    if (read(sub->startfile, buf, EB_SIZE_PAGE) != EB_SIZE_PAGE) {
	eb_error = EB_ERR_FAIL_READ_START;
	return -1;
    }

    /*
     * Reset the index table in `sub'.
     */
    for (i = 0; i < EB_MAX_INDEXES; i++)
	sub->indexes[i] = 0;

    count = eb_uint2(buf);
    if (EB_SIZE_PAGE / 16 - 1 <= count) {
	eb_error = EB_ERR_UNEXP_START;
	return -1;
    }

    /*
     * Load start page numbers of the indexes in the subbook.
     */
    for (i = 0, bufp = buf + 16; i < count; i++, bufp += 16) {
	id = eb_uint1(bufp);
	page = eb_uint4(bufp + 2);
	switch (id) {
	case 0x01:
	    sub->indexes[EB_INDEX_MENU] = page;
	    break;
	case 0x70:
	    sub->indexes[EB_INDEX_ENDWORD_KANA] = page;
	    break;
	case 0x71:
	    sub->indexes[EB_INDEX_ENDWORD_ASIS] = page;
	    break;
	case 0x72:
	    sub->indexes[EB_INDEX_ENDWORD_ALPHA] = page;
	    break;
	case 0x80:
	    sub->indexes[EB_INDEX_KEYWORD] = page;
	    break;
	case 0x90:
	    sub->indexes[EB_INDEX_WORD_KANA] = page;
	    break;
	case 0x91:
	    sub->indexes[EB_INDEX_WORD_ASIS] = page;
	    break;
	case 0x92:
	    sub->indexes[EB_INDEX_WORD_ALPHA] = page;
	    break;
	case 0xf1:
	case 0xf2:
	case 0xf3:
	case 0xf4:
	    if (book->disccode == EB_DISC_EB
		&& sub->fontcount < EB_MAX_FONTS * 2)
		sub->fonts[sub->fontcount++].page = page;
	    break;
	}
    }

    return count;
}


/*
 * For EB/EBG/EBXA, get font information in the current subbook.
 *
 * If succeeded, 0 is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
static int
eb_initialize_eb_fonts(book)
    EB_Book *book;
{
    EB_Font *fnt;
    char buf[16];
    int len;
    int i;

    for (i = 0, fnt = book->subcurrent->fonts;
	 i < book->subcurrent->fontcount; i++, fnt++) {
	/*
	 * Read information from the `START' file.
	 */
	if (lseek(book->subcurrent->startfile, (fnt->page - 1) * EB_SIZE_PAGE,
	    SEEK_SET) < 0) {
	    eb_error = EB_ERR_FAIL_SEEK_START;
	    return -1;
	}
	if (read(book->subcurrent->startfile, buf, 16) != 16) {
	    eb_error = EB_ERR_FAIL_READ_START;
	    return -1;
	}

	/*
	 * Set the information.
	 */
	len = eb_uint2(buf + 12);
	fnt->width = eb_uint1(buf + 8);
	fnt->height = eb_uint1(buf + 9);
	fnt->start = eb_uint2(buf + 10);
	if (book->charcode == EB_CHAR_JISX0208) {
	    fnt->end = fnt->start + ((len / 0x5e) << 8) + (len % 0x5e) - 1;
	    if (0x7e < (fnt->end & 0xff))
		fnt->end += 0xa3;
	} else {
	    fnt->end = fnt->start + ((len / 0xfe) << 8) + (len % 0xfe) - 1;
	    if (0xfe < (fnt->end & 0xff))
		fnt->end += 3;
	}
    }

    return 0;
}


/*
 * For EPWING, get font information in the current subbook.
 *
 * If succeeded, 0 is returned.
 * Otherwise, -1 is returned and `eb_error' is set.
 */
static int
eb_initialize_epwing_fonts(book)
    EB_Book *book;
{
    EB_Font *fnt;
    char filename[PATH_MAX + 1];
    int file;
    char buf[16];
    int len;
    int i;
    
    for (i = 0, fnt = book->subcurrent->fonts;
	 i < book->subcurrent->fontcount; i++, fnt++) {
	/*
	 * Open a font file.
	 */
	sprintf(filename, "%s/%s/%s/%s", book->path,
	    book->subcurrent->directory, EB_DIRNAME_GAIJI, fnt->filename);
	eb_fix_filename(book, filename);
	file = open(filename, O_RDONLY | O_BINARY);
	if (file < 0) {
	    eb_error = EB_ERR_FAIL_OPEN_FONT;
	    return -1;
	}

	/*
	 * Read information.
	 */
	if (read(file, buf, 16) != 16) {
	    eb_error = EB_ERR_FAIL_READ_FONT;
	    return -1;
	}

	/*
	 * Set the information.
	 */
	len = eb_uint2(buf + 12);
	fnt->page = 1;
	fnt->width = eb_uint1(buf + 8);
	fnt->height = eb_uint1(buf + 9);
	fnt->start = eb_uint2(buf + 10);
	if (book->charcode == EB_CHAR_JISX0208) {
	    fnt->end = fnt->start + ((len / 0x5e) << 8) + (len % 0x5e) - 1;
	    if (0x7e < (fnt->end & 0xff))
		fnt->end += 0xa3;
	} else {
	    fnt->end = fnt->start + ((len / 0xfe) << 8) + (len % 0xfe) - 1;
	    if (0xfe < (fnt->end & 0xff))
		fnt->end += 3;
	}

	/*
	 * Close the font file.
	 */
	close(file);
    }

    return 0;
}


