/*
 * Copyright (c) 1997, 1998  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#include "eb.h"
#include "error.h"
#include "font.h"

#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX	MAXPATHLEN
#else /* not MAXPATHLEN */
#define PATH_MAX	1024
#endif /* not MAXPATHLEN */
#endif /* not PATH_MAX */

#ifndef O_BINARY
#define O_BINARY 0
#endif

/*
 * Return the number of subbooks in the book.
 */
int
eb_subbook_count(book)
    EB_Book *book;
{
    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	return -1;
    }

    return book->subcount;
}


/*
 * Make a subbook list in the book.
 */
int
eb_subbook_list(book, list)
    EB_Book *book;
    EB_Subbook_Code *list;
{
    EB_Subbook_Code *lp;
    int i;

    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	return -1;
    }

    for (i = 0, lp = list; i < book->subcount; i++, lp++)
	*lp = i;

    return book->subcount;
}


/*
 * Return a subbook-code of the current subbook.
 */
EB_Subbook_Code
eb_subbook(book)
    EB_Book *book;
{
    /*
     * The current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    return book->subcurrent->code;
}


/*
 * Return a title of the current subbook.
 */
const char *
eb_subbook_title(book)
    EB_Book *book;
{
    /*
     * The current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return NULL;
    }

    return book->subcurrent->title;
}


/*
 * Return a title of the specified subbook `code'.
 */
const char *
eb_subbook_title2(book, code)
    EB_Book *book;
    EB_Subbook_Code code;
{
    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	return NULL;
    }

    /*
     * Check for the subbook-code.
     */
    if (code < 0 || book->subcount <= code) {
	eb_error = EB_ERR_NO_SUCH_SUB;
	return NULL;
    }

    return (book->subbooks + code)->title;
}


/*
 * Return a directory name of the current subbook.
 */
const char *
eb_subbook_directory(book)
    EB_Book *book;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return NULL;
    }

    return book->subcurrent->directory;
}


/*
 * Return a directory name of the specified subbook `code'.
 */
const char *
eb_subbook_directory2(book, code)
    EB_Book *book;
    EB_Subbook_Code code;
{
    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	return NULL;
    }

    /*
     * Check for the subbook-code.
     */
    if (code < 0 || book->subcount <= code) {
	eb_error = EB_ERR_NO_SUCH_SUB;
	return NULL;
    }

    return (book->subbooks + code)->directory;
}


/*
 * Set the subbook `code' as the current subbook.
 */
int
eb_set_subbook(book, code)
    EB_Book *book;
    EB_Subbook_Code code;
{
    char start[PATH_MAX + 1];

    /*
     * The book must have been bound.
     */
    if (book->path == NULL) {
	eb_error = EB_ERR_UNBOUND_BOOK;
	goto failed;
    }

    /*
     * Check for the subbook-code.
     */
    if (code < 0 || book->subcount <= code) {
	eb_error = EB_ERR_NO_SUCH_SUB;
	goto failed;
    }

    /*
     * If the subbook has already been set as the current subbook,
     * there is nothing to be done.
     * Otherwise close the previous subbook.
     */
    if (book->subcurrent != NULL) {
	if (book->subcurrent->code == code)
	    return 0;
	eb_unset_subbook(book);
    }

    /*
     * Open the subbook.
     */
    book->subcurrent = book->subbooks + code;
    if (book->disccode == EB_DISC_EB) {
	sprintf(start, "%s/%s/%s", book->path, book->subcurrent->directory,
	    EB_FNAME_START);
    } else {
	sprintf(start, "%s/%s/%s/%s", book->path, book->subcurrent->directory,
	    EB_DIRNAME_DATA, EB_FNAME_HONMON);
    }
    eb_fix_filename(book, start);
    book->subcurrent->startfile = open(start, O_RDONLY | O_BINARY);
    if (book->subcurrent->startfile < 0) {
	eb_error = EB_ERR_FAIL_OPEN_START;
	book->subcurrent = NULL;
	goto failed;
    }

    /*
     * Initialize the subbook.
     */
    if (eb_initialize_subbook(book) < 0)
	goto failed;

    return 0;

    /*
     * An error occurs...
     */
  failed:
    eb_unset_subbook(book);
    return -1;
}


/*
 * Unset the current subbook.
 */
void
eb_unset_subbook(book)
    EB_Book *book;
{
    /*
     * Close the file of the current subbook.
     */
    if (book->subcurrent != NULL) {
	eb_unset_font(book);
	close(book->subcurrent->startfile);
	book->subcurrent = NULL;
    }

    /*
     * Initialize search and text output status.
     */
    eb_initialize_search();
    eb_initialize_text();
}


