/*
 * Copyright (c) 1997, 1998  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifndef HAVE_MEMCPY
#define memcpy(d, s, n) bcopy((s), (d), (n))
#ifdef __STDC__
void *memchr(const void *, int, size_t);
int memcmp(const void *, const void *, size_t);
void *memmove(void *, const void *, size_t);
void *memset(void *, int, size_t);
#else /* not __STDC__ */
char *memchr();
int memcmp();
char *memmove();
char *memset();
#endif /* not __STDC__ */
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

#include "eb.h"
#include "error.h"
#include "io.h"
#include "text.h"

/*
 * Local macros.
 */
#define CONTENT_TEXT		0
#define CONTENT_HEADING		1

/*
 * Static variables.
 */
static EB_Book_Code bookcode = -1;
static EB_Subbook_Code subcode = -1;
static int method = -1;
static char workbuf[EB_MAXLEN_TEXT_WORK + 1];
static char pagebuf[EB_SIZE_PAGE];
static off_t location;
static int narwflag;
static const EB_Hook *modhook;
static const EB_Hook *refhook;
static char *pagebufp;
static size_t pagerest;
static size_t worklen;
static int charapp;
static int bookeof;
static int textend;

/*
 * The maximum number of arguments for an escape sequence.
 */
#define EB_MAX_ARGV 4

/*
 * Read next page when the length of rest data is less than the value.
 */
#define SIZE_FEW_REST		16

/*
 * Unexported functions.
 */
#ifdef __STDC__
static int eb_text_internal(EB_Book *, EB_Appendix *, const EB_Hookset *,
    char *, size_t);
#else
static int eb_text_internal();
#endif


/*
 * Initialize text processing status.
 */
void
eb_initialize_text()
{
    bookcode = -1;
    subcode = -1;
    method = -1;
}


/*
 * Initialize a hookset.
 */
void
eb_initialize_hookset(hookset)
    EB_Hookset *hookset;
{
    int i;

    for (i = 0; i < EB_NUM_HOOKS; i++) {
	hookset->hooks[i].code = i;
	hookset->hooks[i].function = NULL;
    }
}


/*
 * Set a hook.
 */
int
eb_set_hook(hookset, hook)
    EB_Hookset *hookset;
    const EB_Hook *hook;
{
    if (hook->code < 0 || EB_NUM_HOOKS <= hook->code) {
	eb_error = EB_ERR_NO_SUCH_HOOK;
	return -1;
    }

    if (hook->code != EB_HOOK_NULL)
	hookset->hooks[hook->code].function = hook->function;

    return 0;
}


/*
 * Set a list of hooks.
 */
int
eb_set_hooks(hookset, hook)
    EB_Hookset *hookset;
    const EB_Hook *hook;
{
    const EB_Hook *hk;
    int i;

    for (i = 0, hk = hook;
	 i < EB_NUM_HOOKS && hk->code != EB_HOOK_NULL; i++, hk++) {
	if (hook->code < 0 || EB_NUM_HOOKS <= hook->code) {
	    eb_error = EB_ERR_NO_SUCH_HOOK;
	    return -1;
	}
	hookset->hooks[hk->code].function = hk->function;
    }

    return 0;
}


/*
 * Get a heading in the current subbook in `book'.
 */
int
eb_heading(book, appendix, hookset, position, text, textsize)
    EB_Book *book;
    EB_Appendix *appendix;
    const EB_Hookset *hookset;
    const EB_Position *position;
    char *text;
    size_t textsize;
{
    const EB_Hook *hook;

    /*
     * The current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }
    
    /*
     * Initialize static variables.
     */
    bookcode = book->code;
    subcode = book->subcurrent->code;
    location = (position->page - 1) * EB_SIZE_PAGE + position->offset;
    method = CONTENT_HEADING;
    modhook = hookset->hooks + EB_HOOK_NULL;
    refhook = hookset->hooks + EB_HOOK_NULL;
    pagebufp = pagebuf;
    pagerest = 0;
    narwflag = 0;
    worklen = 0;
    bookeof = 0;
    charapp = 0;
    textend = 0;

    /*
     * Call the function bound to `EB_HOOK_INITIALIZE'.
     */
    hook = hookset->hooks + EB_HOOK_INITIALIZE;
    if (hook->function != NULL && hook->function(book, appendix, NULL,
	EB_HOOK_INITIALIZE, 0, NULL) < 0) {
	eb_error = EB_ERR_HOOK_WORKSPACE;
	return -1;
    }

    return eb_text_internal(book, appendix, hookset, text, textsize);
}


/*
 * Continue to get a heading.
 */
int
eb_next_heading(book, appendix, hookset, text, textsize)
    EB_Book *book;
    EB_Appendix *appendix;
    const EB_Hookset *hookset;
    char *text;
    size_t textsize;
{
    /*
     * The current subbook must been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }
    if (bookcode == -1) {
	eb_error = EB_ERR_NO_PREV_TEXT;
	return -1;
    }
    if (bookcode != book->code) {
	eb_error = EB_ERR_DIFF_BOOK;
	return -1;
    }
    if (subcode != book->subcurrent->code) {
	eb_error = EB_ERR_DIFF_SUBBOOK;
	return -1;
    }
    if (method != CONTENT_HEADING) {
	eb_error = EB_ERR_DIFF_CONTENT;
	return -1;
    }

    if (textend)
	return 0;

    return eb_text_internal(book, appendix, hookset, text, textsize);
}


/*
 * Get text in the current subbook in `book'.
 */
int
eb_text(book, appendix, hookset, position, text, textsize)
    EB_Book *book;
    EB_Appendix *appendix;
    const EB_Hookset *hookset;
    const EB_Position *position;
    char *text;
    size_t textsize;
{
    const EB_Hook *hook;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Initialize static variables.
     */
    bookcode = book->code;
    subcode = book->subcurrent->code;
    location = (position->page - 1) * EB_SIZE_PAGE + position->offset;
    method = CONTENT_TEXT;
    modhook = hookset->hooks + EB_HOOK_NULL;
    refhook = hookset->hooks + EB_HOOK_NULL;
    pagebufp = pagebuf;
    pagerest = 0;
    narwflag = 0;
    worklen = 0;
    bookeof = 0;
    charapp = 0;
    textend = 0;

    /*
     * Call the function bound to `EB_HOOK_INITIALIZE'.
     */
    hook = hookset->hooks + EB_HOOK_INITIALIZE;
    if (hook->function != NULL && hook->function(book, appendix, NULL,
	EB_HOOK_INITIALIZE, 0, NULL) < 0) {
	eb_error = EB_ERR_HOOK_WORKSPACE;
	return -1;
    }

    return eb_text_internal(book, appendix, hookset, text, textsize);
}


/*
 * Continue to get text.
 */
int
eb_next_text(book, appendix, hookset, text, textsize)
    EB_Book *book;
    EB_Appendix *appendix;
    const EB_Hookset *hookset;
    char *text;
    size_t textsize;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    if (bookcode == -1) {
	eb_error = EB_ERR_NO_PREV_TEXT;
	return -1;
    }
    if (bookcode != book->code) {
	eb_error = EB_ERR_DIFF_BOOK;
	return -1;
    }
    if (subcode != book->subcurrent->code) {
	eb_error = EB_ERR_DIFF_SUBBOOK;
	return -1;
    }
    if (method != CONTENT_TEXT) {
	eb_error = EB_ERR_DIFF_CONTENT;
	return -1;
    }

    if (textend)
	return 0;

    return eb_text_internal(book, appendix, hookset, text, textsize);
}


/*
 * Get text or heading.
 */
static int
eb_text_internal(book, appendix, hookset, text, textsize)
    EB_Book *book;
    EB_Appendix *appendix;
    const EB_Hookset *hookset;
    char *text;
    size_t textsize;
{
    char *textp = text;
    unsigned char c;
    size_t textrest = textsize;
    const EB_Hook *hook;
    int step;
    int argv[EB_MAX_ARGV];
    int argc;

    /*
     * If unprocessed data are rest in `textrest', copy them to `workbuf'.
     */
    if (0 < worklen) {
	if (textrest < worklen) {
	    *text = '\0';
	    return 0;
	}
	strcpy(textp, workbuf);
	textp += worklen;
	textrest -= worklen;
	worklen = 0;
    }

    for (;;) {
	step = 0;
	workbuf[0] = '\0';
	argc = 1;

	/*
	 * If reached to the near of buffer's end, read next page.
	 */
	if (pagerest < SIZE_FEW_REST && !bookeof) {
	    ssize_t len;

	    memcpy(pagebuf, pagebufp, pagerest);
	    if (lseek(book->subcurrent->startfile, location, SEEK_SET) == -1) {
		eb_error = EB_ERR_FAIL_SEEK_START;
		return -1;
	    }
	    len = read(book->subcurrent->startfile, pagebuf + pagerest,
		EB_SIZE_PAGE - pagerest);
	    if (len < 0) {
		eb_error = EB_ERR_FAIL_READ_START;
		return -1;
	    } else if (len != EB_SIZE_PAGE - pagerest)
		bookeof = 1;
	    location += len;
	    pagebufp = pagebuf;
	    pagerest += len;
	}

	/*
	 * Get 1 byte from the buffer.
	 */
	if (pagerest < 1) {
	    eb_error = EB_ERR_UNEXP_START;
	    return -1;
	}
	c = eb_uint1(pagebufp);

	if (c == 0x1f) {
	    const EB_Hook *ctlhook = hookset->hooks + EB_HOOK_NULL;

	    if (pagerest < 2) {
		eb_error = EB_ERR_UNEXP_START;
		return -1;
	    }
	    step = 2;
	    argv[0] = eb_uint2(pagebufp);

	    /*
	     * Is is an escape sequences.
	     */
	    switch (eb_uint1(pagebufp + 1)) {
	    case 0x02:
		/* beginning of text */
		break;

	    case 0x03:
		/* end of text */
		goto at_end;

	    case 0x04:
		/* beginning of NARROW */
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_NARROW;
		narwflag = 1;
		break;

	    case 0x05:
		/* end of NARROW */
		ctlhook = hookset->hooks + EB_HOOK_END_NARROW;
		narwflag = 0;
		break;

	    case 0x06:
		/* beginning of subscript */
		modhook = hookset->hooks + EB_HOOK_SUBSCRIPT;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_SUBSCRIPT;
		break;

	    case 0x07:
		/* end of subscript */
		modhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_SUBSCRIPT;
		break;

	    case 0x09:
		/* set indent */
		if (pagerest < step + 2) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		ctlhook = hookset->hooks + EB_HOOK_SET_INDENT;
		step += 2;
		argc = 2;
		argv[1] = eb_uint2(pagebufp + 2);

		hook = hookset->hooks + EB_HOOK_STOPCODE;
		if (charapp && hook->function != NULL && hook->function(book,
		    appendix, workbuf, EB_HOOK_STOPCODE, argc, argv) < 0)
		    goto at_end;
		break;

	    case 0x0a:
		/* newline */
		if (method == CONTENT_HEADING) {
		    textend = 1;
		    goto at_end;
		}
		*workbuf = '\n';
		*(workbuf + 1) = '\0';
		ctlhook = hookset->hooks + EB_HOOK_NEWLINE;
		break;

	    case 0x0e:
		/* beginning of superscript */
		modhook = hookset->hooks + EB_HOOK_SUPERSCRIPT;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_SUPERSCRIPT;
		break;

	    case 0x0f:
		/* end of superscript */
		modhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_SUPERSCRIPT;
		break;

	    case 0x10:
		/* beginning of newline prohibition */
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_NO_NEWLINE;
		break;

	    case 0x11:
		/* end of newline prohibition */
		ctlhook = hookset->hooks + EB_HOOK_END_NO_NEWLINE;
		break;

	    case 0x12:
		/* beginning of emphasis */
		modhook = hookset->hooks + EB_HOOK_EMPHASIS;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_EMPHASIS;
		break;

	    case 0x13:
		/* end of emphasis */
		modhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_EMPHASIS;
		break;

	    case 0x32:
		/* beginning of an anchor of the picture */
		modhook = hookset->hooks + EB_HOOK_PICTURE;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_PICTURE;
		break;

	    case 0x33:
		/* beginning of an anchor of the sound */
		refhook = hookset->hooks + EB_HOOK_SOUND;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_PICTURE;
		break;

	    case 0x41:
		/* beginning of the keyword */
		if (pagerest < step + 2) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		refhook = hookset->hooks + EB_HOOK_KEYWORD;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_KEYWORD;
		step += 2;
		argc = 2;
		argv[1] = eb_uint2(pagebufp + 2);

		hook = hookset->hooks + EB_HOOK_STOPCODE;
		if (charapp && hook->function != NULL && hook->function(book,
		    appendix, workbuf, EB_HOOK_STOPCODE, argc, argv) < 0)
		    goto at_end;
		break;

	    case 0x42:
		/* beginning of the reference */
		if (pagerest < step + 2) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		if (eb_uint1(pagebufp + 2) == 0x00)
		    step += 2;
		refhook = hookset->hooks + EB_HOOK_REFERENCE;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_REFERENCE;
		break;

	    case 0x43:
		/* beginning of an entry of the menu */
		refhook = hookset->hooks + EB_HOOK_MENU;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_MENU;
		break;

	    case 0x44:
		/* beginning of an anchor of the picture (old style?) */
		if (pagerest < step + 10) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		modhook = hookset->hooks + EB_HOOK_PICTURE;
		ctlhook = hookset->hooks + EB_HOOK_BEGIN_PICTURE;
		step += 10;
		break;

	    case 0x45:
		/* prefix of sound or picuture */
		if (pagerest < step + 2) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		if (eb_uint1(pagebufp + 2) != 0x1f) {
		    argc = 2;
		    argv[1] = eb_bcd4(pagebufp + 2);
		    step += 2;
		}
		break;

	    case 0x52:
		/* end of the picture */
		if (pagerest < step + 6) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		refhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_SOUND;
		step += 6;
		argc = 3;
		argv[1] = eb_bcd4(pagebufp + 2);
		argv[2] = eb_bcd2(pagebufp + 6);
		break;

	    case 0x53:
		/* end of the sound */
		if (pagerest < step + 8) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		refhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_PICTURE;
		step += 8;
		break;

	    case 0x61:
		/* end of the keyword */
		refhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_KEYWORD;
		break;

	    case 0x62:
		/* end of the reference */
		if (pagerest < step + 6) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		refhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_REFERENCE;
		step += 6;
		argc = 3;
		argv[1] = eb_bcd4(pagebufp + 2);
		argv[2] = eb_bcd2(pagebufp + 6);
		break;

	    case 0x63:
		/* end of an entry of the menu */
		if (pagerest < step + 6) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		refhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_MENU;
		step += 6;
		argc = 3;
		argv[1] = eb_bcd4(pagebufp + 2);
		argv[2] = eb_bcd2(pagebufp + 6);
		break;

	    case 0x64:
		/* end of the picture (old style?) */
		if (pagerest < step + 6) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}
		refhook = hookset->hooks + EB_HOOK_NULL;
		ctlhook = hookset->hooks + EB_HOOK_END_PICTURE;
		step += 6;
		argc = 3;
		argv[1] = eb_bcd4(pagebufp + 2);
		argv[2] = eb_bcd2(pagebufp + 6);
		break;

	    case 0x1a:
	    case 0x1b:
	    case 0x1c:
	    case 0x1d:
	    case 0x1e:
	    case 0x1f:
	    case 0xe0:
		/* emphasis; described in JIS X 4081-1996 */
		if (book->disccode == EB_DISC_EB) {
		    if (pagerest < step + 1) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }
		} else {
		    if (pagerest < step + 2) {
			eb_error = EB_ERR_UNEXP_START;
			return -1;
		    }
		    step += 2;
		}
		ctlhook = hookset->hooks + EB_HOOK_UNKNOWN_SEQUENCE;
		break;

	    default:
		ctlhook = hookset->hooks + EB_HOOK_UNKNOWN_SEQUENCE;
		break;
	    }

	    if (ctlhook->function != NULL && ctlhook->function(book, appendix,
		workbuf, ctlhook->code, argc, argv) < 0) {
		eb_error = EB_ERR_HOOK_WORKSPACE;
		return -1;
	    }
	} else if (book->charcode == EB_CHAR_JISX0208) {
	    /*
	     * The book is written in JIS X 0208.
	     */
	    charapp = 1;
	    step = 2;
	    argv[0] = eb_uint2(pagebufp);

	    if (pagerest < 2) {
		eb_error = EB_ERR_UNEXP_START;
		return -1;
	    }

	    if (0x20 < c && c < 0x7f) {
		/*
		 * This is a JIS X 0208 KANJI character.
		 */
		*workbuf = (c | 0x80);
		*(workbuf + 1) = (eb_uint1(pagebufp + 1) | 0x80);
		*(workbuf + 2) = '\0';
		if (narwflag) {
		    hook = hookset->hooks + EB_HOOK_NARROW;
		    if (hook->function != NULL && hook->function(book,
			appendix, workbuf, EB_HOOK_NARROW, 0, argv) < 0) {
			eb_error = EB_ERR_HOOK_WORKSPACE;
			return -1;
		    }
		}
		hook = hookset->hooks + EB_HOOK_JISX0208;
		if (hook->function != NULL && hook->function(book, appendix,
		    workbuf, EB_HOOK_JISX0208, argc, argv) < 0) {
		    eb_error = EB_ERR_HOOK_WORKSPACE;
		    return -1;
		}
	    } else {
		/*
		 * This is a local character.
		 */
		*workbuf = c;
		*(workbuf + 1) = eb_uint1(pagebufp + 1);
		*(workbuf + 2) = '\0';
		if (narwflag) {
		    hook = hookset->hooks + EB_HOOK_NARROW_FONT;
		    if (hook->function != NULL && hook->function(book,
			appendix, workbuf, EB_HOOK_NARROW_FONT, argc, argv)
			< 0) {
			eb_error = EB_ERR_HOOK_WORKSPACE;
			return -1;
		    }
		} else {
		    hook = hookset->hooks + EB_HOOK_WIDE_FONT;
		    if (hook->function != NULL && hook->function(book,
			appendix, workbuf, EB_HOOK_WIDE_FONT, argc, argv)
			< 0) {
			eb_error = EB_ERR_HOOK_WORKSPACE;
			return -1;
		    }
		}
	    }

	    if (modhook->function != NULL && modhook->function(book, appendix,
		workbuf, modhook->code, 0, argv) < 0) {
		eb_error = EB_ERR_HOOK_WORKSPACE;
		return -1;
	    }
	    if (refhook->function != NULL && refhook->function(book, appendix,
		workbuf, refhook->code, 0, argv) < 0) {
		eb_error = EB_ERR_HOOK_WORKSPACE;
		return -1;
	    }
	} else {
	    /*
	     * The book is mainly written in ISO 8859 1.
	     */
	    charapp = 1;

	    if ((0x20 <= c && c < 0x7f) || (0xa0 <= c && c < 0xff)) {
		/*
		 * This is an ISO 8859 1 character.
		 */
		step = 1;
		argv[0] = eb_uint1(pagebufp);
		*workbuf = c;
		*(workbuf + 1) = '\0';
		hook = hookset->hooks + EB_HOOK_ISO8859_1;
		if (hook->function != NULL && hook->function(book, appendix,
		    workbuf, EB_HOOK_ISO8859_1, argc, argv) < 0) {
		    eb_error = EB_ERR_HOOK_WORKSPACE;
		    return -1;
		}
	    } else {
		/*
		 * This is a local character.
		 */
		if (pagerest < 2) {
		    eb_error = EB_ERR_UNEXP_START;
		    return -1;
		}

		step = 2;
		argv[0] = eb_uint2(pagebufp);
		*workbuf = c;
		*(workbuf + 1) = eb_uint1(pagebufp + 1);
		*(workbuf + 2) = '\0';
		hook = hookset->hooks + EB_HOOK_NARROW_FONT;
		if (hook->function != NULL && hook->function(book, appendix,
		    workbuf, EB_HOOK_NARROW_FONT, argc, argv) < 0) {
		    eb_error = EB_ERR_HOOK_WORKSPACE;
		    return -1;
		}
	    }

	    if (modhook->function != NULL && modhook->function(book, appendix,
		workbuf, modhook->code, 0, argv) < 0) {
		eb_error = EB_ERR_HOOK_WORKSPACE;
		return -1;
	    }

	    if (refhook->function != NULL && refhook->function(book, appendix,
		workbuf, refhook->code, 0, argv) < 0) {
		eb_error = EB_ERR_HOOK_WORKSPACE;
		return -1;
	    }
	}

	worklen = strlen(workbuf);
	if (textrest < worklen)
	    goto at_end;

	strcpy(textp, workbuf);
	textp += worklen;
	textrest -= worklen;
	pagebufp += step;
	pagerest -= step;
	worklen = 0;
    }

    /*
     * At end.
     */
  at_end:
    *textp = '\0';
    return (textsize - textrest);
}


