/*
 * Copyright (c) 1997, 1998  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <syslog.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "eb/eb.h"
#include "eb/error.h"
#include "eb/font.h"

#include "fakelog.h"
#include "getopt.h"

#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif /* HAVE_STRCHR */

/*
 * Unexported functions.
 */
#ifdef __STDC__
static int output_information(const char *);
static void output_version(void);
static void output_help(void);
static void output_try_help(void);
#else
static int output_information();
static void output_version();
static void output_help();
static void output_try_help();
#endif

/*
 * Long style options.
 */
static struct option longopts[] = {
  {"help",    no_argument, NULL, 'h'},
  {"version", no_argument, NULL, 'v'},
  {NULL, 0, NULL, 0}
};

/*
 * Generic name of the program.
 */
const char *program_name = "ebfont";

/*
 * Program version.
 */
const char *program_version = VERSION;

/*
 * Actual program name. (argv[0])
 */
static const char *invoked_name;


int
main(argc, argv)
    int argc;
    char *argv[];
{
    int ch;
    char *bookdir;

    invoked_name = argv[0];
    set_fakelog(FAKELOG_TO_STDERR, LOG_ERR);

    /*
     * Parse command line options.
     */
    for (;;) {
	ch = getopt_long(argc, argv, "hv", longopts, NULL);
	if (ch == EOF)
	    break;
	switch (ch) {
	case 'h':
	    /*
	     * Option `-h'.  Display help message, then exit.
	     */
	    output_help();
	    exit(0);

	case 'v':
	    /*
	     * Option `-v'.  Display version number, then exit.
	     */
	    output_version();
	    exit(0);

	default:
	    output_try_help();
	    exit(1);
	}
    }

    /*
     * Check the number of rest arguments.
     */
    if (1 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\n", invoked_name);
	output_try_help();
	exit(1);
    }

    /*
     * Output information about the book.
     */
    if (argc == optind)
	bookdir = ".";
    else
	bookdir = argv[optind];

    if (output_information(bookdir) < 0)
	exit(1);

    return 0;
}


/*
 * Output information about the book at `path'.
 */
static int
output_information(path)
    const char *path;
{
    EB_Book book;
    EB_Disc_Code disc;
    EB_Character_Code charcode;
    EB_Subbook_Code sublist[EB_MAX_SUBBOOKS];
    EB_Font_Code fontlist[EB_MAX_FONTS];
    const char *str;
    int start, end;
    int subcount;
    int fontcount;
    int i, j;

    /*
     * Start to use a book.
     */
    if (eb_bind(&book, path) < 0)
	goto failed;

    /*
     * Output disc type.
     */
    disc = eb_disc_type(&book);
    if (disc < 0)
	goto failed;
    if (disc == EB_DISC_EB)
	printf("disc type: EB/EBG/EBXA\n");
    else
	printf("disc type: EPWING\n");

    /*
     * Output character code.
     */
    charcode = eb_character_code(&book);
    if (charcode < 0)
	goto failed;
    if (charcode == EB_CHAR_JISX0208)
	printf("character code: JIS X 0208\n");
    else 
	printf("character code: ISO 8859-1\n");

    /*
     * Output the number of subbooks in the book.
     */
    subcount = eb_subbook_list(&book, sublist);
    if (subcount < 0)
	goto failed;
    printf("the number of subbooks: %d\n\n", subcount);

    /*
     * Output information about each subbook.
     */
    for (i = 0; i < subcount; i++) {
	printf("subbook %d:\n", i + 1);

	/*
	 * Output a title of the subbook.
	 */
	str = eb_subbook_title2(&book, sublist[i]);
	if (str == NULL)
	    goto failed;
	printf("  title: %s\n", str);

	/*
	 * Output a directory name of the subbook.
	 */
	str = eb_subbook_directory2(&book, sublist[i]);
	if (str == NULL)
	    goto failed;
	printf("  directory: %s\n", str);

	/*
	 * Set the current subbook to `i'.
	 */
	if (eb_set_subbook(&book, sublist[i]) < 0) {
	    fprintf(stderr, "%s: %s\n\n", invoked_name, eb_error_message());
	    continue;
	}

	/*
	 * Output supported methods.
	 */
	printf("  search methods: ");
	if (eb_have_word_search(&book))
	    fputs("word ", stdout);
	if (eb_have_endword_search(&book))
	    fputs("endword ", stdout);
	if (eb_have_keyword_search(&book))
	    fputs("keyword ", stdout);
	if (eb_have_multi_search(&book))
	    fputs("multi ", stdout);
	if (eb_have_graphic_search(&book))
	    fputs("graphic ", stdout);
	if (eb_have_menu_search(&book))
	    fputs("menu ", stdout);
	fputc('\n', stdout);

	/*
	 * Output a font list.
	 */
	fputs("  font sizes: ", stdout);
	fontcount = eb_font_list(&book, fontlist);
	for (j = 0; j < fontcount; j++)
	    printf("%d ", (int)fontlist[j]);
	fputc('\n', stdout);

	/*
	 * Output characters information about the font.
	 */
	fputs("  narrow font characters: ", stdout);
	if (eb_have_narrow_font(&book)) {
	    if (eb_set_font(&book, fontlist[0]) < 0)
		goto failed;
	    start = eb_narrow_font_start(&book);
	    if (start < 0)
		goto failed;
	    end = eb_narrow_font_end(&book);
	    if (end < 0)
		goto failed;
	    printf("0x%04x -- 0x%04x", start, end);
	}
	fputc('\n', stdout);

	fputs("  wide font characters: ", stdout);
	if (eb_have_wide_font(&book)) {
	    if (eb_set_font(&book, fontlist[0]) < 0)
		goto failed;
	    start = eb_wide_font_start(&book);
	    if (start < 0)
		goto failed;
	    end = eb_wide_font_end(&book);
	    if (end < 0)
		goto failed;
	    printf("0x%04x -- 0x%04x", start, end);
	}
	fputs("\n\n", stdout);
    }
    fflush(stdout);

    /*
     * End to use the book.
     */
    eb_clear(&book);

    return 0;

    /*
     * An error occurs...
     */
  failed:
    fputc('\n', stdout);
    fflush(stdout);
    fprintf(stderr, "%s: %s\n", invoked_name, eb_error_message());
    fflush(stderr);
    eb_clear(&book);

    return -1;
}


/*
 * Output version number to stdandard error.
 */
static void
output_version()
{
    fprintf(stderr, "%s version %s\n", program_name, program_version);
    fflush(stderr);
}


/*
 * Output help message to stdandard error.
 */
static void
output_help()
{
    fprintf(stderr, "Usage: %s [option...] [book-directory]\n",
	program_name);
    fprintf(stderr, "Options:\n");
    fprintf(stderr,
        "  -h, --help                 display this help, then exit\n");
    fprintf(stderr,
	"  -v, --version              display version number, then exit\n");
    fprintf(stderr, "\nArgument:\n");
    fprintf(stderr,
	"  book-directory             top directory of a CD-ROM book\n");
    fprintf(stderr,
	"                             (default: .)\n");
    fflush(stderr);
}


/*
 * Output ``try ...'' message to standard error.
 */
static void
output_try_help()
{
    fprintf(stderr, "try `%s --help' for more information\n", invoked_name);
    fflush(stderr);
}

