/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <syslog.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <eb/error.h>
#include <eb/eb.h>
#include <eb/io.h>
#include <eb/text.h>

#include "ioall.h"
#include "linebuf.h"
#include "permission.h"
#include "wildcard.h"

#include "ndtpd.h"

#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif /* HAVE_STRCHR */

#ifndef HAVE_STRTOL
#ifdef __STDC__
long strtol(const char *, char **, int);
#else /* not __STDC__ */
long strtol();
#endif /* not __STDC__ */
#endif /* not HAVE_STRTOL */


/*
 * Unexported functions.
 */
#ifdef __STDC__
static int command_L(int, const char *);
static int command_Q(int, const char *);
static int command_A(int, const char *);
static int command_u(int, const char *);
static int command_P(int, const char *);
static int command_F(int, const char *);
static int command_S(int, const char *);
static int command_I(int, const char *);
static int command_v(int, const char *);
static int command_T(int, const char *);
static int command_t(int, const char *);
static int book_number_to_name(char *, int);
static void reverse_word_A(char *);
static void reverse_word_K(char *);

#else /* not __STDC__ */

static int command_L();
static int command_Q();
static int command_A();
static int command_u();
static int command_P();
static int command_F();
static int command_S();
static int command_I();
static int command_v();
static int command_T();
static int command_t();
static int book_number_to_name();
static void reverse_word_A();
static void reverse_word_K();
#endif /* not __STDC__ */


/*
 * NDTP command table.
 */
typedef struct {
    char name;
#ifdef __STDC__
    int (*function)(int, const char *);
#else
    int (*function)();
#endif
} NDTP_Command;

static const NDTP_Command ndtp_commands[] = {
    {'L', command_L},
    {'Q', command_Q},
    {'A', command_A},
    {'u', command_u},
    {'P', command_P},
    {'F', command_F},
    {'S', command_S},
    {'I', command_I},
    {'v', command_v},
    {'T', command_T},
    {'t', command_t},
    {'\0', NULL}
};


/*
 * Communicate with a client by NDTP.
 */
#define MAXLEN_SHORTBUF 4

int
ndtp_main()
{
    char buf[MAXLEN_NDTP_LINE + 1];
    const NDTP_Command *cmd;
    ssize_t len;

    reset_line_buffer(&line_buffer);

    for (;;) {
	/*
	 * Set an alarm.
	 */
	alarm(idle_timeout);
	
	/*
	 * Read a line from the client.
	 */
	len = read_line_buffer(&line_buffer, buf, accepted_file);
	if (len < 0)
	    goto die;
	else if (len == 0)
	    continue;

	/*
	 * Check for the line length.
	 */
	if (*(buf + len) != '\0') {
	    *(buf + MAXLEN_SHORTBUF) = '\0';
	    syslog(LOG_NOTICE, "the line too long: line=%s..., user=%s, \
host=%s(%s)",
		buf, client_user, client_hostname, client_address);
	    if (write_all(accepted_file, "$&\n", 3) < 0)
		return -1;
	    continue;
	}
	syslog(LOG_DEBUG, "debug: received line: %s", buf);

	/*
	 * Scan the command table.
	 */
	for (cmd = ndtp_commands; cmd->name != '\0'; cmd++) {
	    if (cmd->name == *buf)
		break;
	}
	if (cmd->name == '\0') {
	    *(buf + 1) = '\0';
	    syslog(LOG_ERR, "unknown command: command=%s, user=%s, \
host=%s(%s)",
		buf, client_user, client_hostname, client_address);
	    if (write_all(accepted_file, "$?\n", 3) < 0)
		goto die;

	    continue;
	}

	/*
	 * Set an alarm, again.
	 */
	alarm(idle_timeout);

	/*
	 * Dispatch.
	 */
	if ((cmd->function)(*buf, buf + 1) < 0)
	    goto die;

	if (*buf == 'Q')
	    break;
    }

    alarm(0);
    return 0;

  die:
    alarm(0);
    return -1;
}


/*
 * Command `L'.  Change the current book/subbook.
 * It takes an argument which represents a book/subbook name.
 */
static int
command_L(command, arg)
    int command;
    const char *arg;
{
    char bookname[MAXLEN_NDTP_LINE + 1];
    char *subname;
    char realarg[MAXLEN_BOOKNAME + 1 + EB_MAXLEN_BASENAME + 1];
    char *slash;
    char *endp;
    int number;

    /*
     * Unset the current book.
     */
    unset_current_book();

    /*
     * If `arg' is `nodict', unset the current book/subbook.
     */
    if (strcmp(arg, "nodict") == 0) {
	if (write_all(accepted_file, "$*\n", 3) < 0)
	    return -1;
	syslog(LOG_NOTICE, "succeeded: command=L, book=nodict, user=%s, \
host=%s(%s)",
	    client_user, client_hostname, client_address);
	return 0;
    }

    /*
     * For dserver 2.0 compatibility, check for aliases `eiwa', `waei',
     * and `kojien'.
     */
    if (strcmp(arg, "eiwa") == 0 && alias_eiwa[0] != '\0')
	strcpy(realarg, alias_eiwa);
    else if (strcmp(arg, "waei") == 0 && alias_waei[0] != '\0')
	strcpy(realarg, alias_waei);
    else if (strcmp(arg, "kojien") == 0 && alias_kojien[0] != '\0')
	strcpy(realarg, alias_kojien);
    else
	strcpy(realarg, arg);

    /*
     * If the argument doesn't contain a slash (`/'), it must be a book
     * number.
     */
    slash = strchr(realarg, '/');
    if (slash == NULL) {
	number = strtol(realarg, &endp, 10);

	/*
	 * Convert a book number to a book/subbook name.
	 */
	if (!isdigit(realarg[0]) || *endp != '\0'
	    || book_number_to_name(realarg, number) < 0) {
	    syslog(LOG_ERR,
		"unknown book: command=L, book=%s, user=%s, host=%s(%s)",
		realarg, client_user, client_hostname, client_address);
	    if (write_all(accepted_file, "$&\n", 3) < 0)
		return -1;
	    return 0;
	}

	/*
	 * If `arg' is `0', unset the current book/subbook.
	 */
	if (number == 0) {
	    if (write_all(accepted_file, "$*\n", 3) < 0)
		return -1;
	    syslog(LOG_NOTICE, "succeeded: command=L, book=nodict, user=%s, \
host=%s(%s)",
		client_user, client_hostname, client_address);
	    return 0;
	}

	slash = strchr(realarg, '/');
    }

    /*
     * Get book and subbook name separated by a slash (`/').
     */
    strncpy(bookname, realarg, (int)(slash - realarg));
    *(bookname + (int)(slash - realarg)) = '\0';
    subname = slash + 1;

    /*
     * Check for the book name.
     */
    if (set_current_book(bookname) < 0) {
	syslog(LOG_ERR, "unknown book: command=L, book=%s, user=%s, \
host=%s(%s)",
	    realarg, client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$&\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Check for access permission to the book.
     */
    if (!current_book->permflag) {
	syslog(LOG_NOTICE, "denied: command=L, book=%s, user=%s, host=%s(%s)",
	    realarg, client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$&\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Check for the subbook name.
     */
    if (set_current_subbook(subname) < 0) {
	syslog(LOG_ERR, "%s: command=L, book=%s, user=%s, host=%s(%s)",
	    eb_error_message(), realarg, client_user, client_hostname,
	    client_address);
	if (write_all(accepted_file, "$&\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Send a response.
     */
    if (write_all(accepted_file, "$*\n", 3) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=L, book=%s, user=%s, host=%s(%s)",
	realarg, client_user, client_hostname, client_address);
    return 0;
}


/*
 * Command `Q'.  Quit.
 */
static int
command_Q(command, arg)
    int command;
    const char *arg;
{
    syslog(LOG_NOTICE, "succeeded: command=Q, user=%s, host=%s(%s)",
	client_user, client_hostname, client_address);
    return 0;
}


/*
 * Command `A'.  Get access permission.
 * The command is not supported.  It always sends the `succeeded' response.
 */
static int
command_A(command, arg)
    int command;
    const char *arg;
{
    if (write_all(accepted_file, "$A\n", 3) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=A, user=%s, host=%s(%s)",
	client_user, client_hostname, client_address);
    return 0;
}


/*
 * Command `U'.  List current users.
 * The command is not supported.  It always sends an empty list.
 */
static int
command_u(command, arg)
    int command;
    const char *arg;
{
    if (write_all(accepted_file, "$U\n$$\n", 6) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=U, user=%s, host=%s(%s)",
	client_user, client_hostname, client_address);
    return 0;
}


#define NUM_HIT_LIST	50

/*
 * Command `P'.  Search a word.
 * It takes two arguments without a separater.
 * The first argument is a character which represents an index type.
 * The second argument represents a word to search.
 */
static int
command_P(command, arg)
    int command;
    const char *arg;
{
#ifdef __STDC__
    int (*search)(EB_Book *, EB_Hit *, const char *, int);
    int (*next)(EB_Book *, EB_Hit *, int);
#else
    int (*search)();
    int (*next)();
#endif
    EB_Hit hitlist[NUM_HIT_LIST];
    EB_Hit *hit;
    char message[MAXLEN_STRING + 1];
    char word[EB_MAXLEN_WORD + 1];
    char method;
    char *asterisk;
    ssize_t headlen;
    size_t wordlen;
    int count;
    int n;
    int i;

    /*
     * Check for the current book.
     */
    if (current_book == NULL) {
	syslog(LOG_ERR, "no current book: command=P, user=%s, host=%s(%s)",
	    client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$N\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Set a word to search and a search method.
     */
    wordlen = strlen(arg) - 1;
    if (EB_MAXLEN_WORD < wordlen) {
	syslog(LOG_ERR, "%s: command=P, user=%s, host=%s(%s)", 
	    eb_error_message2(EB_ERR_TOO_LONG_WORD), client_user,
	    client_hostname, client_address);
	if (write_all(accepted_file, "$&\n", 3) < 0)
	    return -1;
	return 0;
    }
    if (*arg == '\0') {
	syslog(LOG_ERR, "empty method name: command=P, user=%s, host=%s(%s)",
	    client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$?\n", 3) < 0)
	    return -1;
	return 0;
    }
    strcpy(word, arg + 1);
    method = *arg;

    /*
     * Check for the search method.
     */
    if (method != 'a' && method != 'k' && method != 'A' && method != 'K') {
	syslog(LOG_ERR, "bad search method: command=P, method=%s, \
user=%s, host=%s(%s)",
	    method, client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$?\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Remove `*' in the tail of the word and choose a function.
     */
    asterisk = strrchr(word, '*');
    if (asterisk != NULL && *(asterisk + 1) == '\0') {
	*asterisk = '\0';
	if (method == 'a' || method == 'k') {
	    search = eb_search_word;
	    next = eb_next_word;
	} else if (method == 'A' || method == 'K') {
	    search = eb_search_endword;
	    next = eb_next_endword;
	}
    } else {
	search = eb_search_exactword;
	next = eb_next_exactword;
    }

    /*
     * Reverse the word when endword search.
     */
    if (method == 'A')
	reverse_word_A(word);
    else if (method == 'K')
	reverse_word_K(word);

    /*
     * Search the word.
     */
    count = 0;
    n = search(&current_book->book, hitlist, word, NUM_HIT_LIST);
    if (write_all(accepted_file, "$0\n", 3) < 0)
	return -1;

    while (0 < n && (count < max_hits || max_hits == 0)) {
	for (i = 0, hit = hitlist; i < n; i++, count++, hit++) {
	    /*
	     * Get heading.
	     */
	    headlen = eb_heading(&current_book->book, &current_book->appendix,
		&text_hookset, &hit->heading, message, MAXLEN_STRING);
	    if (headlen < 0) {
		syslog(LOG_ERR, "%s: command=P, book=%s/%s, method=%c, \
word=%s, user=%s, host=%s(%s)",
		    eb_error_message(), current_book->name,
		    current_book->subname, method, arg + 1, client_user,
		    client_hostname, client_address);
		break;
	    }
	    if (headlen == 0)
		continue;

	    /*
	     * Output the heading and text position.
	     */
	    if (write_all(accepted_file, message, strlen(message)) < 0)
		return -1;
	    sprintf(message, "\n%lx:%lx\n", (long)hit->text.page,
		(long)hit->text.offset);
	    if (write_all(accepted_file, message, strlen(message)) < 0)
		return -1;
	}

	/*
	 * Continue searching.
	 */
	n = next(&current_book->book, hitlist, NUM_HIT_LIST);
    }

    if (n < 0) {
	syslog(LOG_ERR, "%s: command=P, book=%s/%s, method=%c, word=%s, \
user=%s, host=%s(%s)",
	    eb_error_message(), current_book->name, current_book->subname,
	    method, arg + 1, client_user, client_hostname, client_address);
    }
    if (write_all(accepted_file, "$$\n", 3) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=P, book=%s/%s, method=%c, \
word=%s, matches=%d, user=%s, host=%s(%s)",
	current_book->name, current_book->subname, method, arg + 1, count,
	client_user, client_hostname, client_address);
    return 0;
}


/*
 * Command `F'.  Send a page.
 * It takes an argument which represents a page to send.
 * The page number must be a hexadecimal integer.
 */
static int
command_F(command, arg)
    int command;
    const char *arg;
{
    EB_Position pos;
    char page[EB_SIZE_PAGE];
    char *endp;

    /*
     * Check for the current book.
     */
    if (current_book == NULL) {
	syslog(LOG_ERR, "no current book: command=F, user=%s, host=%s(%s)",
	    client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$N\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Parse an page number.
     */
    pos.offset = 0;
    pos.page = strtol(arg, &endp, 16);
    if (!isxdigit(*arg) || *endp != '\0') {
	syslog(LOG_ERR, "bad page number: command=F, page=%s, user=%s, \
host=%s(%s)",
	    arg, client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$?\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Read the page.
     */
    if (eb_seek(&current_book->book, &pos) < 0) {
	syslog(LOG_ERR, "%s: command=F, book=%s/%s, page=%ld, user=%s, \
host=%s(%s)",
	    eb_error_message(), current_book->name, current_book->subname,
	    (long)pos.page, client_hostname, client_address);
	if (write_all(accepted_file, "$&\n", 3) < 0)
	    return -1;
	return 0;
    }
    if (eb_read(&current_book->book, page, EB_SIZE_PAGE) != EB_SIZE_PAGE) {
	syslog(LOG_ERR, "%s: command=F, book=%s/%s, page=%ld, user=%s, \
host=%s(%s)",
	    eb_error_message(), current_book->name, current_book->subname,
	    (long)pos.page, client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$&\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Send the page to the client.
     */
    if (write_all(accepted_file, "$F", 2) < 0)
	return -1;
    if (write_all(accepted_file, page, EB_SIZE_PAGE) < 0)
	return -1;
    if (write_all(accepted_file, "\n", 1) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=F, book=%s/%s, page=%ld, \
user=%s, host=%s(%s)",
	current_book->name, current_book->subname, (long)pos.page,
	client_user, client_hostname, client_address);
    return 0;
}


/*
 * Command `S' and `s'.  Send text.
 * It takes two arguments; page number and offset separeted by a colon (`:').
 * Both the page number and offset must be hexadecimal integers.
 */
static int
command_S(command, arg)
    int command;
    const char *arg;
{
    EB_Position pos;
    char message[EB_SIZE_PAGE];
    char *endp;
    char *offsetp;
    ssize_t len;

    /*
     * Check for the current book.
     */
    if (current_book == NULL) {
	syslog(LOG_ERR, "no current book: command=S, user=%s, host=%s(%s)",
	    client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$N\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Parse a page number.
     */
    pos.page = strtol(arg, &endp, 16);
    if (!isxdigit(*arg) || *endp != ':') {
	syslog(LOG_ERR, "bad position: command=S, position=%s, user=%s, \
host=%s(%s)",
	    arg, client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$?\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Parse an offset number.
     */
    offsetp = endp + 1;
    pos.offset = strtol(offsetp, &endp, 16);
    if (!isxdigit(*offsetp) || *endp != '\0') {
	syslog(LOG_ERR, "bad position: command=S, position=%s, user=%s, \
host=%s(%s)",
	    arg, client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$?\n", 3) < 0)
	    return -1;
	return 0;
    }
    len = EB_SIZE_PAGE - pos.offset;

    /*
     * Read text and send them to the client.
     */
    if (write_all(accepted_file, "$1\n", 3) < 0)
	return -1;
    if (0 < eb_text(&current_book->book, &current_book->appendix, 
	&text_hookset, &pos, message, EB_SIZE_PAGE - 1)) {
	if (write_all(accepted_file, message, strlen(message)) < 0)
	    return -1;
        while (0 < eb_next_text(&current_book->book, &current_book->appendix, 
	    &text_hookset, message, EB_SIZE_PAGE - 1)) {
	    if (write_all(accepted_file, message, strlen(message)) < 0)
		return -1;
	}
    }
    if (write_all(accepted_file, "\n$$\n", 4) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=S, book=%s/%s, position=%s, \
user=%s, host=%s(%s)",
	current_book->name, current_book->subname, arg, client_user,
	client_hostname, client_address);
    return 0;
}


/*
 * Command `I'.  Send an index list.
 * The page numbers in the table are dummy except for menu search.
 */
static int
command_I(command, arg)
    int command;
    const char *arg;
{
    char message[MAXLEN_STRING + 1];
    EB_Position pos;

    /*
     * Check for the current book.
     */
    if (current_book == NULL) {
	syslog(LOG_ERR, "no current book: command=I, user=%s, host=%s(%s)",
	    client_user, client_hostname, client_address);
	if (write_all(accepted_file, "$N\n", 3) < 0)
	    return -1;
	return 0;
    }

    /*
     * Send response data.
     */
    if (write_all(accepted_file, "$I\n", 3) < 0)
	return -1;
    if (eb_have_word_search(&current_book->book)) {
	strcpy(message, "IA 1\nIK 1\n");
	if (write_all(accepted_file, message, strlen(message)) < 0)
	    return -1;
    }
    if (eb_have_endword_search(&current_book->book)) {
	strcpy(message, "BA 1\nBK 1\n");
	if (write_all(accepted_file, message, strlen(message)) < 0)
	    return -1;
    }
    if (0 <= eb_search_menu(&current_book->book, &pos)) {
	sprintf(message, "HA %lx\n", (long)pos.page);
	if (write_all(accepted_file, message, strlen(message)) < 0)
	    return -1;
    }

    if (write_all(accepted_file, "$$\n", 3) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=I, book=%s/%s, user=%s, \
host=%s(%s)",
	current_book->name, current_book->subname, client_user,
	client_hostname, client_address);
    return 0;
}


/*
 * Command `v'.  Output version number of the server.
 */
static int
command_v(command, arg)
    int command;
    const char *arg;
{
    char message[MAXLEN_STRING + 1];

    sprintf(message, "$v%s version %s on %s\n", program_name, program_version,
	server_name);
    if (write_all(accepted_file, message, strlen(message)) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=v, user=%s, host=%s(%s)",
	client_user, client_hostname, client_address);
    return 0;
}


/*
 * Command `T'.  Output a book list.
 */
static int
command_T(command, arg)
    int command;
    const char *arg;
{
    EB_Subbook_Code sublist[EB_MAX_SUBBOOKS];
    Book *book;
    char message[MAXLEN_STRING + 1];
    const char *title;
    int count;
    int i, j;

    /*
     * Send a book list.
     */
    for (book = book_registry, i = 1;
	 book != NULL; book = (Book *)book->next) {
	count = eb_subbook_list(&book->book, sublist);
	if (count < 0 || !book->permflag)
	    continue;
	for (j = 0; j < count; j++, i++) {
	    title = eb_subbook_title2(&book->book, sublist[j]);
	    if (title == NULL)
		title = "(not available)";
	    sprintf(message, "%d\t%s\n", i, title);
	    if (write_all(accepted_file, message, strlen(message)) < 0)
		return -1;
	}
    }
    if (write_all(accepted_file, "$*\n", 3) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=T, user=%s, host=%s(%s)",
	client_user, client_hostname, client_address);
    return 0;
}


/*
 * Command `t'.  List current users and books.
 * The command is not supported.  It always sends an empty list.
 */
static int
command_t(command, arg)
    int command;
    const char *arg;
{
    EB_Subbook_Code sublist[EB_MAX_SUBBOOKS];
    Book *book;
    char message[MAXLEN_STRING + 1];
    const char *title, *dir;
    int count;
    int i, j;

    /*
     * Send a book list.
     */
    for (book = book_registry, i = 1;
	 book != NULL; book = (Book *)book->next) {
	count = eb_subbook_list(&book->book, sublist);
	if (count < 0 || !book->permflag)
	    continue;
	for (j = 0; j < count; j++, i++) {
	    dir = eb_subbook_directory2(&book->book, sublist[j]);
	    if (dir == NULL)
		dir = "*";
	    title = eb_subbook_title2(&book->book, sublist[j]);
	    if (title == NULL)
		title = "(not available)";
	    sprintf(message, "%d\t%s\t%s/%s\t%d\t%d\t%d\n", i, title,
		book->name, dir, 0, book->max_clients, idle_timeout);
	    if (write_all(accepted_file, message, strlen(message)) < 0)
		return -1;
	}
    }
    if (write_all(accepted_file, "$*\n", 3) < 0)
	return -1;

    syslog(LOG_NOTICE, "succeeded: command=t, user=%s, host=%s(%s)",
	client_user, client_hostname, client_address);
    return 0;
}


/*
 * Reverse a word for ENDWORD SEARCH.
 * 
 * `word' is a word to reverse.  It must be an alphabetic word.
 * The reversed word is also put into `word'.
 */
static void
reverse_word_A(word)
    char *word;
{
    char *p1, *p2;
    int len;
    char c;

    len = strlen(word);
    if (len == 0)
	return;
    for (p1 = word, p2 = word + len - 1; p1 < p2; p1++, p2--) {
	c = *p1;
	*p1 = *p2;
	*p2 = c;
    }
}


/*
 * Reverse a word for ENDWORD SEARCH.
 * 
 * `word' is a word to reverse.  It must be a KANA word.
 * The reversed word is also put into `word'.
 */
static void
reverse_word_K(word)
    char *word;
{
    char *p1, *p2;
    int len;
    char c;

    len = strlen(word);
    if (len % 2 == 1) {
	*(word + len - 1) = '\0';
	len--;
    }
    if (len < 2)
	return;
    
    for (p1 = word, p2 = word + len - 2; p1 < p2; p1 += 2, p2 -= 2) {
	c = *p1;
	*p1 = *p2;
	*p2 = c;
	c = *(p1 + 1);
	*(p1 + 1) = *(p2 + 1);
	*(p2 + 1) = c;
    }
}


/*
 * Get a book number corresponding to a book number.
 */
static int
book_number_to_name(name, number)
    char *name;
    int number;
{
    EB_Subbook_Code sublist[EB_MAX_SUBBOOKS];
    Book *book;
    const char *dir;
    int count;
    int i;

    if (number == 0) {
	strcpy(name, "nodict");
	return 0;
    }

    for (book = book_registry, i = 1;
	 book != NULL && i <= number; book = (Book *)book->next) {
	count = eb_subbook_list(&book->book, sublist);
	if (count < 0 || !book->permflag)
	    continue;
	if (number < i + count) {
	    dir = eb_subbook_directory2(&book->book, sublist[number - i]);
	    if (dir == NULL)
		return -1;
	    sprintf(name, "%s/%s", book->name, dir);
	    return 0;
	}
	i += count;
    }

    return -1;
}


