/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#if defined(HAVE_SYS_WAIT_H) || defined(HAVE_UNION_WAIT)
#include <sys/wait.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif /* HAVE_STRCHR */

#ifndef HAVE_GETCWD
#define getcwd(d,n) getwd(d)
#endif

#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX        MAXPATHLEN
#else /* not MAXPATHLEN */
#define PATH_MAX        1024
#endif /* not MAXPATHLEN */
#endif /* not PATH_MAX */

#ifdef HAVE_UNION_WAIT
#ifndef WTERMSIG
#define WTERMSIG(x)     ((x).w_termsig)
#endif
#ifndef WCOREDUMP
#define WCOREDUMP(x)    ((x).w_coredump)
#endif
#ifndef WEXITSTATUS
#define WEXITSTATUS(x)  ((x).w_retcode)
#endif
#ifndef WIFSIGNALED
#define WIFSIGNALED(x)  (WTERMSIG(x) != 0)
#endif
#ifndef WIFEXITED
#define WIFEXITED(x)    (WTERMSIG(x) == 0)
#endif
#else /* not HAVE_UNION_WAIT */
#ifndef WTERMSIG
#define WTERMSIG(x) ((x) & 0x7f)
#endif
#ifndef WCOREDUMP
#define WCOREDUMP(x) ((x) & 0x80)
#endif
#ifndef WEXITSTATUS
#define WEXITSTATUS(x) (((x) >> 8) & 0xff)
#endif
#ifndef WIFSIGNALED
#define WIFSIGNALED(x) (WTERMSIG (x) != 0)
#endif
#ifndef WIFEXITED
#define WIFEXITED(x) (WTERMSIG (x) == 0)
#endif
#endif  /* not HAVE_UNION_WAIT */


#include "confutil.h"
#include "daemon.h"
#include "filename.h"
#include "getopt.h"
#include "hostname.h"
#include "ident.h"
#include "linebuf.h"
#include "logpid.h"
#include "openmax.h"
#include "permission.h"
#include "privilege.h"
#include "serverport.h"
#include "signame.h"
#include "wildcard.h"

#include "ndtpd.h"

/*
 * Unexported functions.
 */
#ifdef __STDC__
static void standalone_main(void);
static void inetd_main(void);
static void output_version(void);
static void output_help(void);
static void output_try_help(void);
static RETSIGTYPE abort_parent(int);
static RETSIGTYPE abort_child(int);
static RETSIGTYPE terminate_parent(int);
static RETSIGTYPE terminate_child(int);
static RETSIGTYPE restart_parent(int);
static RETSIGTYPE wait_child(int);
#else /* not __STDC__ */
static void standalone_main();
static void inetd_main();
static void output_version();
static void output_help();
static void output_try_help();
static RETSIGTYPE abort_parent();
static RETSIGTYPE abort_child();
static RETSIGTYPE terminate_parent();
static RETSIGTYPE terminate_child();
static RETSIGTYPE restart_parent();
static RETSIGTYPE wait_child();
#endif /* not __STDC__ */

/*
 * Environment buffer for setjmp() or sigsetjmp().
 */
#ifdef HAVE_SIGSETJMP
sigjmp_buf jump_buffer;
#else
jmp_buf jump_buffer;
#endif

/*
 * Program name.
 */
#define PROGRAM_NAME	"ndtpd"

/*
 * Long style options.
 */
static struct option longopts[] = {
    {"configuration-file", required_argument, NULL, 'c'},
    {"help",               no_argument,       NULL, 'h'},
    {"inetd",              no_argument,       NULL, 'i'},
    {"version",            no_argument,       NULL, 'v'},
    {NULL, 0, NULL, 0}
};

/*
 * The flag is set once the server is started
 */
static int once_started;

/*
 * Process group ID of the server.
 */
pid_t process_group_id;


int
main(argc, argv)
    int argc;
    char *argv[];
{
    int ch;

    /*
     * Set program name and version.
     */
    invoked_name = argv[0];
    program_name = PROGRAM_NAME;
    program_version = PROGRAM_VERSION;

    /*
     * Set umask.
     */
    umask(SERVER_UMASK);

    /*
     * Open syslog.
     * If LOG_DAEMON is not defined, it assumes old style syslog
     * It doesn't have facility.
     */
    syslog_facility = DEFAULT_SYSLOG_FACILITY;
#ifdef LOG_DAEMON
    openlog(program_name, LOG_NDELAY | LOG_PID, syslog_facility);
#else
    openlog(program_name, LOG_NDELAY | LOG_PID);
#endif
    syslog(LOG_DEBUG, "debug: open syslog");

    /*
     * Initialize variables.
     */
    listening_file = -1;
    server_mode = SERVER_MODE_STANDALONE;
    strcpy(configuration_filename, DEFAULT_CONFIG_FILENAME);
    if (initialize_line_buffer(&line_buffer, MAXLEN_NDTP_LINE) < 0) {
	syslog(LOG_ERR, "memory exhausted");
	goto die;
    }
    initialize_permission(&permissions);
    initialize_permission(&identifications);
    initialize_book_registry();
    initialize_text_hookset();

    /*
     * Parse command line options.
     */
    for (;;) {
	ch = getopt_long(argc, argv, "c:hiv", longopts, NULL);
	if (ch == EOF)
	    break;
	switch (ch) {
	case 'c':
	    /*
	     * Option `-c file'.  Specify the configuration filename.
	     */
	    if (MAXLEN_FILENAME < strlen(optarg)) {
		fprintf(stderr, "%s: too long configuration filename\n",
		    invoked_name);
		goto die;
	    }
	    strcpy(configuration_filename, optarg);
	    if (canonicalize_filename(configuration_filename) < 0)
		goto die;
	    break;

	case 'h':
	    /*
	     * Option `-h'.  Help.
	     */
	    output_help();
	    exit(0);

	case 'i':
	    /*
	     * Option `-i'.  Inetd mode.
	     */
	    server_mode = SERVER_MODE_INETD;
	    break;

	case 'v':
	    /*
	     * Option `-v'.  Show version and exit immediately.
	     */
	    output_version();
	    exit(0);

	default:
	    output_try_help();
	    exit(1);
	}
    }

    /*
     * Check for the number of rest arguments.
     */
    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\n", invoked_name);
	output_try_help();
	exit(1);
    }

    /*
     * Main routine for each server mode.
     */
    if (server_mode == SERVER_MODE_INETD)
	inetd_main();
    else
	standalone_main();

    return 0;

    /*
     * A fatal error occurrs...
     */
  die:
    syslog(LOG_CRIT, "the server dies");
    exit(1);
}


/*
 * Main routine for standalone mode.
 */
static void
standalone_main()
{
    struct sockaddr_in addr;
    int len;
    int fd;
    int pid;

    /*
     * Initalize variables.
     */
    once_started = 0;

    /*
     * Close unwanted files.
     */
    closelog();
    for (fd = get_open_max() - 1; 0 <= fd; fd--)
	close(fd);
#ifdef LOG_DAEMON
    openlog(program_name, LOG_NDELAY | LOG_PID, syslog_facility);
#else
    openlog(program_name, LOG_NDELAY | LOG_PID);
#endif
    syslog(LOG_DEBUG, "debug: reopen syslog");

    /*
     * Set the re-start point for SIGHUP.
     */
#ifdef HAVE_SIGSETJMP
    sigsetjmp(jump_buffer, 1);
#else
    setjmp(jump_buffer);
#endif
    signal(SIGHUP, SIG_IGN);

    /*
     * Read a configuration file.
     */
    if (read_configuration(configuration_filename, configuration_table) < 0) {
	syslog(LOG_ERR, "configuration failure");
	goto die_parent;
    }

    /*
     * Reopen syslog.
     */
    closelog();
#ifdef LOG_DAEMON
    openlog(program_name, LOG_NDELAY | LOG_PID, syslog_facility);
#else
    openlog(program_name, LOG_NDELAY | LOG_PID);
#endif

    /*
     * Run the process as daemon.
     */
    if (!once_started) {
	if (daemonize() < 0)
	    goto die_parent;
#ifdef GETPGRP_VOID
	process_group_id = getpgrp();
#else
	process_group_id = getpgrp(getpid());
#endif
    }

    /*
     * Change the current working directory.
     */
    if (chdir(work_path) < 0) {
	syslog(LOG_ERR, "cannot change the directory, %m: %s", work_path);
	goto die_parent;
    }

    /*
     * Kill child processes.
     */
    if (once_started) {
#ifdef HAVE_KILLPG
	killpg(process_group_id, SIGHUP);
#else
	kill(-process_group_id, SIGHUP);
#endif
	syslog(LOG_DEBUG, "debug: kill all child processes");
    }
    connection_count = 0;

    /*
     * Shutdown old connections if the server port is changed.
     */
    if (once_started && listening_port != old_listening_port) {
	if (shutdown(listening_file, 2) < 0) {
	    syslog(LOG_ERR, "shutdown() failed, %m");
	    listening_file = -1;
	    goto die_parent;
	}
	listening_file = -1;
	syslog(LOG_DEBUG, "debug: shutdown the old socket: %d/tcp",
	    listening_port);
    }

    /*
     * Bind and listen to the server port. 
     */
    if (listening_file < 0) {
	listening_file = set_server_port(listening_port, LISTENING_BACKLOG);
	if (listening_file < 0)
	    goto die_parent;
	old_listening_port = listening_port;
    }

    /*
     * Set UID and GID.
     */
    if (set_privilege(user_id, group_id) < 0)
	goto die_parent;

    /*
     * Unlink an old PID file if exists.
     */
    if (once_started)
	remove_pid_file(old_pid_filename);
    strcpy(old_pid_filename, pid_filename);

    /*
     * Log pid to the file.
     */
    if (log_pid_file(pid_filename) < 0)
	goto die_parent;

    /*
     * Set signal handlers.
     */
    signal(SIGBUS, abort_parent);
    signal(SIGSEGV, abort_parent);
    signal(SIGINT, terminate_parent);
    signal(SIGTERM, terminate_parent);
    signal(SIGCHLD, wait_child);
    signal(SIGHUP, restart_parent);

    /*
     * Set `once_started' flag.
     */
    if (once_started)
	syslog(LOG_NOTICE, "server restarted");
    else
	syslog(LOG_NOTICE, "server started");
    once_started = 1;

    /*
     * Main loop at the parent process.
     */
    for (;;) {
	/*
	 * The connection counter reaches to its limit, pause until a
	 * child die.
	 */
	if (max_clients <= connection_count)
	    pause();

	/*
	 * Wait a connection from a client.
	 */ 
	if (wait_server_port(listening_file) < 0)
	    goto die_parent;

	/*
	 * Accept a connection.
	 */
	len = sizeof(struct sockaddr_in);
	accepted_file = accept(listening_file, (struct sockaddr *)&addr, &len);
	if (accepted_file < 0) {
	    syslog(LOG_ERR, "accept() failed, %m");
	    continue;
	}
	syslog(LOG_DEBUG, "debug: accept the connection");

	/*
	 * Fork.  The child process talk to the client.
	 */
	pid = fork();
	if (pid < 0) {
	    syslog(LOG_ERR, "fork() failed, %m");
	    continue;
	} else if (pid == 0) {
	    syslog(LOG_DEBUG, "debug: forked");
	    break;
	}

	/*
	 * Close an accepted socket (at the parent process).
	 */
	connection_count++;
	if (close(accepted_file) < 0) {
	    syslog(LOG_ERR, "cannot close the accepted file");
	    goto die_parent;
	}
    }

    /*
     * Set signal handlers.
     */
    signal(SIGBUS, abort_child);
    signal(SIGSEGV, abort_child);
    signal(SIGINT, terminate_child);
    signal(SIGTERM, terminate_child);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGHUP, terminate_child);
    signal(SIGPIPE, SIG_IGN);

    /*
     * Close the file to listen to the server port.
     */
    close(listening_file);

    /*
     * Get the client hostname and address.
     */
    if (identify_remote_host(accepted_file, client_hostname, client_address)
	< 0) {
	syslog(LOG_ERR, "connection denied: host=%s(%s)", client_hostname,
	    client_address);
	if (shutdown(accepted_file, 2) < 0)
	    syslog(LOG_ERR, "shutdown() failed, %m");
	exit(1);
    }

    /*
     * Identify a remote user by Identification protocol, if needed.
     */
    if (test_permission(&identifications, client_hostname, client_address, 
	match_wildcard))
	identify_user(accepted_file, client_user, identification_timeout);
    else
	strcpy(client_user, UNKNOWN_USER);

    /*
     * Check for the access permission to the client.
     */
    if (!test_permission(&permissions, client_hostname, client_address,
	match_wildcard)) {
	syslog(LOG_ERR, "connection denied: user=%s, host=%s(%s)",
	    client_user, client_hostname, client_address);
	if (shutdown(accepted_file, 2) < 0) {
	    syslog(LOG_ERR, "shutdown() failed, %m");
	    accepted_file = -1;
	    goto die_child;
	}
	exit(0);
    }

    syslog(LOG_NOTICE, "connected: user=%s, host=%s(%s)", client_user,
	client_hostname, client_address);

    /*
     * Set the signal handler, SIGALRM.
     */
    signal(SIGALRM, terminate_child);

    /*
     * Set book permissions to the current client.
     */
    set_book_permissions();

    /*
     * Communicate with the client.
     */
    if (ndtp_main() < 0)
	goto die_child;
    if (shutdown(accepted_file, 2) < 0) {
	syslog(LOG_ERR, "shutdown() failed, %m");
	accepted_file = -1;
	goto die_child;
    }
    syslog(LOG_ERR, "the child server process exits");
    exit(0);

    /*
     * A fatal error occurrs on a parent...
     */
    /*
     * Shutdown the connection. 
     */
  die_parent:
    if (0 <= listening_file) {
	if (shutdown(listening_file, 2) < 0) 
	    syslog(LOG_ERR, "shutdown() failed, %m");
	else
	    syslog(LOG_DEBUG, "debug: shutdown the connection");
    }

    /*
     * Unlink the old PID file if exists.
     */
    if (once_started)
	remove_pid_file(old_pid_filename);

    /*
     * Kill child processes.
     */
    if (once_started) {
#ifdef HAVE_KILLPG
	killpg(process_group_id, SIGHUP);
#else
	kill(-process_group_id, SIGHUP);
#endif
    }

    syslog(LOG_CRIT, "the server dies");
    exit(1);

    /*
     * A fatal error occurrs on a child...
     */
    /*
     * Shutdown the connection. 
     */
  die_child:
    if (0 <= accepted_file) {
	if (shutdown(accepted_file, 2) < 0) {
	    syslog(LOG_ERR, "shutdown() failed, %m: user=%s, host=%s(%s)",
		client_user, client_hostname, client_address);
	    accepted_file = -1;
	} else {
	    syslog(LOG_NOTICE, "shutdown the connection: user=%s, host=%s(%s)",
		client_user, client_hostname, client_address);
	}
    }
    
    syslog(LOG_CRIT, "the child server process dies");
    exit(1);
}


/*
 * Main routine for inetd mode.
 */
static void
inetd_main()
{
    accepted_file = 0;

    /*
     * Set signal handlers.
     */
    signal(SIGBUS, abort_child);
    signal(SIGSEGV, abort_child);
    signal(SIGINT, terminate_child);
    signal(SIGTERM, terminate_child);
    signal(SIGHUP, terminate_child);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGALRM, terminate_child);

    /*
     * Read a configuration file.
     */
    if (read_configuration(configuration_filename, configuration_table) < 0) {
	syslog(LOG_ERR, "configuration failure");
	goto die;
    }

    /*
     * Reopen syslog.
     */
    closelog();
#ifdef LOG_DAEMON
    openlog(program_name, LOG_NDELAY | LOG_PID, syslog_facility);
#else
    openlog(program_name, LOG_NDELAY | LOG_PID);
#endif

    /*
     * Set UID and GID.
     */
    if (set_privilege(user_id, group_id) < 0) {
	syslog(LOG_ERR,
	    "cannot set owner and group of the process");
	goto die;
    }

    /*
     * Get the client hostname and address.
     */
    if (identify_remote_host(accepted_file, client_hostname, client_address)
	< 0) {
	syslog(LOG_ERR, "connection denied: host=%s(%s)", client_hostname,
	    client_address);
	goto die;
    }

    /*
     * Identify a remote user by Identification protocol, if needed.
     */
    if (test_permission(&identifications, client_hostname, client_address, 
	match_wildcard))
	identify_user(accepted_file, client_user, identification_timeout);
    else
	strcpy(client_user, UNKNOWN_USER);

    /*
     * Check for the access permission to the client.
     */
    if (!test_permission(&permissions, client_hostname, client_address,
	match_wildcard)) {
	syslog(LOG_ERR, "connection denied: user=%s, host=%s(%s)",
	    client_user, client_hostname, client_address);
	if (shutdown(accepted_file, 2) < 0)
	    syslog(LOG_ERR, "shutdown() failed, %m");
	exit(0);
    }

    /*
     * Set book permissions to the current client.
     */
    set_book_permissions();

    /*
     * Communicate with the client.
     */
    ndtp_main();
    exit(0);

    /*
     * A fatal error occurrs...
     */
  die:
    if (shutdown(accepted_file, 2) < 0)
	syslog(LOG_ERR, "shutdown() failed, %m");

    syslog(LOG_CRIT, "the server dies");
    exit(1);
}


/*
 * Output version number to standard error.
 */
static void
output_version()
{
    fprintf(stderr, "%s version %s\n", program_name, program_version);
    fflush(stderr);
}


/*
 * Output usage to standard error.
 */
static void
output_help()
{
    fprintf(stderr, "Usage: %s [option...]\n", program_name);
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -c FILE, --configuration-file FILE\n");
    fprintf(stderr, "                             specify a configuration file\n");
    fprintf(stderr, "                             (default: %s)\n",
	DEFAULT_CONFIG_FILENAME);
    fprintf(stderr, "  -h, --help                 display this help, then exit\n");
    fprintf(stderr, "  -i, --inetd                run as a child of inetd\n");
    fprintf(stderr, "  -v, --version              display version number, then exit\n");
    fprintf(stderr, "\nDefault value in a configuration file:\n");
    fprintf(stderr, "  work-path                  (default: %s)\n", 
	DEFAULT_WORK_PATH);
    fflush(stderr);
}


/*
 * Output ``try ...'' message to standard error.
 */
static void
output_try_help()
{
    fprintf(stderr, "try `%s --help' for more information\n", invoked_name);
    fflush(stderr);
}


/*
 * Signal handler for SIGBUS and SIGSEGV. (for parent)
 */
static RETSIGTYPE
abort_parent(sig)
    int sig;
{
    alarm(0);
    syslog(LOG_ERR, "the server process aborts, receives SIG%s",
	signal_name(sig));

    abort();

#ifndef RETSIGTYPE_VOID
    return 0;
#endif
}


/*
 * Signal handler for SIGBUS and SIGSEGV. (for child)
 */
static RETSIGTYPE
abort_child(sig)
    int sig;
{
    alarm(0);
    syslog(LOG_ERR, "the child server process aborts, receives SIG%s",
	signal_name(sig));
    abort();

#ifndef RETSIGTYPE_VOID
    return 0;
#endif
}


/*
 * Signal handler for SIGTERM, SIGINT. (for parent)
 */
static RETSIGTYPE
terminate_parent(sig)
    int sig;
{
    syslog(LOG_ERR, "the server process exits, receives SIG%s",
	signal_name(sig));

    /*
     * Kill child processes.
     */
    signal(SIGHUP, SIG_IGN);
    if (once_started && server_mode == SERVER_MODE_STANDALONE) {
#ifdef HAVE_KILLPG
	killpg(process_group_id, SIGHUP);
#else
	kill(-process_group_id, SIGHUP);
#endif
    }

    /*
     * Shutdown a server port.
     */
    if (server_mode == SERVER_MODE_STANDALONE) {
	if (shutdown(listening_file, 2) < 0)
	    syslog(LOG_ERR, "shutdown() failed, %m");
    }

    /*
     * Remove a PID file.
     */
    if (server_mode == SERVER_MODE_STANDALONE)
	remove_pid_file(old_pid_filename);

    exit(1);

#ifndef RETSIGTYPE_VOID
    return 0;
#endif
}


/*
 * Signal handler for SIGHUP, SIGINT, SIGALRM, and SIGTERM. (for child)
 */
static RETSIGTYPE
terminate_child(sig)
    int sig;
{
    alarm(0);
    syslog(LOG_ERR, "the child server process exits, receives SIG%s",
	signal_name(sig));

    if (shutdown(accepted_file, 2) < 0)
	syslog(LOG_ERR, "shutdown() failed, %m");

    exit(0);

#ifndef RETSIGTYPE_VOID
    return 0;
#endif
}


/*
 * Signal handler for SIGCHLD.
 */
static RETSIGTYPE
wait_child(sig)
    int sig;
{
#ifdef HAVE_UNION_WAIT
    union wait status;
#else
    int status;
#endif

    syslog(LOG_NOTICE,
	"child server process stops or exits, receives SIGCHLD");

#ifdef HAVE_WAITPID
    while (0 < waitpid(-1, &status, WNOHANG)) {
	if (WIFEXITED(status) || WIFSIGNALED(status))
	    connection_count--;
    }
#else /* not HAVE WAITPID */
    while (0 < wait3(&status, WNOHANG, NULL)) {
	if (WIFEXITED(status) || WIFSIGNALED(status))
	    connection_count--;
    }
#endif /* not HAVE WAITPID */

    /*
     * Set a signal handler again.
     */
    signal(SIGCHLD, wait_child);

#ifndef RETSIGTYPE_VOID
    return 0;
#endif
}


/*
 * Signal handler for SIGHUP.
 */
static RETSIGTYPE
restart_parent(sig)
    int sig;
{
    alarm(0);

    syslog(LOG_ERR, "the server restarts, receives SIGHUP");

#ifdef HAVE_SIGSETJMP
    siglongjmp(jump_buffer, 1);
#else
    longjmp(jump_buffer, 1);
#endif

#ifndef RETSIGTYPE_VOID
    return 0;
#endif
}
